from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q

COLUMN_CRITERIA_NONE = 0
COLUMN_CRITERIA_EQUAL = 1
COLUMN_CRITERIA_NOT_EQUAL = 2
COLUMN_CRITERIA_LESS_THAN = 3
COLUMN_CRITERIA_LESS_THAN_OR_EQUAL = 4
COLUMN_CRITERIA_GREATER_THAN = 5
COLUMN_CRITERIA_GREATER_THAN_OR_EQUAL = 6
COLUMN_CRITERIA_CONTAINS = 7
COLUMN_CRITERIA_STARTS_WITH = 8
COLUMN_CRITERIA_ENDS_WITH = 9
COLUMN_CRITERIA_TRUE = 10
COLUMN_CRITERIA_FALSE = 11
COLUMN_CRITERIA_NULL = 12
COLUMN_CRITERIA_NOT_NULL = 13

COLUMN_TEXT_CRITERIA_SET = [
    COLUMN_CRITERIA_NONE,
    COLUMN_CRITERIA_EQUAL,
    COLUMN_CRITERIA_NOT_EQUAL,
    COLUMN_CRITERIA_CONTAINS,
    COLUMN_CRITERIA_STARTS_WITH,
    COLUMN_CRITERIA_ENDS_WITH,
    COLUMN_CRITERIA_NULL,
    COLUMN_CRITERIA_NOT_NULL
]

COLUMN_DATE_CRITERIA_SET = [
    COLUMN_CRITERIA_NONE,
    COLUMN_CRITERIA_EQUAL,
    COLUMN_CRITERIA_NOT_EQUAL,
    COLUMN_CRITERIA_CONTAINS,
    COLUMN_CRITERIA_STARTS_WITH,
    COLUMN_CRITERIA_ENDS_WITH,
    COLUMN_CRITERIA_NULL,
    COLUMN_CRITERIA_NOT_NULL
]

COLUMN_NUMBER_CRITERIA_SET = [
    COLUMN_CRITERIA_NONE,
    COLUMN_CRITERIA_EQUAL,
    COLUMN_CRITERIA_NOT_EQUAL,
    COLUMN_CRITERIA_LESS_THAN,
    COLUMN_CRITERIA_LESS_THAN_OR_EQUAL,
    COLUMN_CRITERIA_GREATER_THAN,
    COLUMN_CRITERIA_GREATER_THAN_OR_EQUAL,
    COLUMN_CRITERIA_NULL,
    COLUMN_CRITERIA_NOT_NULL
]

COLUMN_BOOLEAN_CRITERIA_SET = [
    COLUMN_CRITERIA_NONE,
    COLUMN_CRITERIA_TRUE,
    COLUMN_CRITERIA_FALSE
]

COLUMN_CRITERIA_SET = {
    COLUMN_CRITERIA_NONE: "(aucun)",
    COLUMN_CRITERIA_EQUAL: "égal à",
    COLUMN_CRITERIA_NOT_EQUAL: "différent de",
    COLUMN_CRITERIA_LESS_THAN: "plus petit que",
    COLUMN_CRITERIA_LESS_THAN_OR_EQUAL: "plus petit ou égal",
    COLUMN_CRITERIA_GREATER_THAN: "plus grand que",
    COLUMN_CRITERIA_GREATER_THAN_OR_EQUAL: "plus grand ou égal",
    COLUMN_CRITERIA_CONTAINS: "contient",
    COLUMN_CRITERIA_STARTS_WITH: "commence par",
    COLUMN_CRITERIA_ENDS_WITH: "termine par",
    COLUMN_CRITERIA_TRUE: "vrai",
    COLUMN_CRITERIA_FALSE: "faux",
    COLUMN_CRITERIA_NULL: "vide",
    COLUMN_CRITERIA_NOT_NULL: "non vide",
}


class Log(models.Model):
    user: User = models.ForeignKey(User, on_delete=models.DO_NOTHING, null=True, related_name='actor')
    ref = models.CharField(max_length=64, default='')
    model = models.CharField(max_length=64, default='')
    date = models.DateTimeField(auto_now_add=True, editable=True, null=True, blank=True)
    diff = models.TextField(default='')
    action = models.CharField(max_length=64, default='')
    transaction = models.CharField(max_length=64, default='', null=True)

    objects = models.Manager()

    def __str__(self):
        return str(self.id)

    def author(self):
        return self.user.username if self.user else '(admin)'


class LogItem(models.Model):
    log = models.ForeignKey(Log, on_delete=models.DO_NOTHING, null=True, related_name='items')
    key = models.CharField(max_length=128, default='')
    old = models.CharField(max_length=128, default='', null=True)
    new = models.CharField(max_length=128, default='', null=True)

    objects = models.Manager()

    def __str__(self):
        return str(self.id)


class GridTable:
    def __init__(self, label, fields, manager, table, url, reports=[]):
        self.label = label
        self.fields = fields
        self.manager = manager
        self.table = table
        self.url = url
        self.reports = reports


class Grid(models.Model):
    name = models.CharField(max_length=255, default='', null=True, blank=True)
    table = models.CharField(max_length=255, default='', null=True, blank=True)
    comment = models.CharField(max_length=255, default='', null=True, blank=True)
    filter_by = models.CharField(max_length=512, default='', null=True, blank=True)
    group_by = models.BooleanField(default=False)
    distinct = models.BooleanField(default=False)
    show_on_home = models.BooleanField(default=False)
    objects = models.Manager()

    def __str__(self):
        return str(self.name)

    def table_as_str(self):
        return settings.GRID_TABLES[self.table].label if self.table in settings.GRID_TABLES else ''

    def fields(self):
        return settings.GRID_TABLES[self.table].fields if self.table in settings.GRID_TABLES else None


class Column(models.Model):
    grid: Grid = models.ForeignKey(Grid, on_delete=models.DO_NOTHING, null=True)

    label = models.CharField(max_length=255, default='', null=True, blank=True)
    value = models.CharField(max_length=255, default='', null=True, blank=True)
    criteria = models.IntegerField(default=COLUMN_CRITERIA_EQUAL)
    width = models.IntegerField(default=80, null=True, blank=True)
    order = models.IntegerField(default=0)
    hidden = models.BooleanField(default=False)

    objects = models.Manager()

    def __str__(self):
        return str(self.id)

    def label_as_str(self):
        fields = self.grid.fields() if self.grid else None
        for field in fields:
            if field.key == self.label:
                return field.label
        return self.label

    def label_with_operator(self, op):
        return str(self.label) + '__' + op

    def criteria_as_str(self):
        return COLUMN_CRITERIA_SET[self.criteria]

    def filter(self, objects):
        try:
            if self.criteria == COLUMN_CRITERIA_NOT_EQUAL:
                return objects.exclude(Q(**{self.label: self.value}))
            elif self.criteria == COLUMN_CRITERIA_EQUAL:
                return objects.filter(Q(**{self.label: self.value}))
            elif self.criteria == COLUMN_CRITERIA_LESS_THAN:
                return objects.filter(Q(**{self.label_with_operator('lt'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_LESS_THAN_OR_EQUAL:
                return objects.filter(Q(**{self.label_with_operator('lte'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_GREATER_THAN:
                return objects.filter(Q(**{self.label_with_operator('gt'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_GREATER_THAN_OR_EQUAL:
                return objects.filter(Q(**{self.label_with_operator('gte'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_CONTAINS:
                return objects.filter(Q(**{self.label_with_operator('contains'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_STARTS_WITH:
                return objects.filter(Q(**{self.label_with_operator('startswith'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_ENDS_WITH:
                return objects.filter(Q(**{self.label_with_operator('endswith'): self.value}))
            elif self.criteria == COLUMN_CRITERIA_TRUE:
                return objects.filter(Q(**{self.label: True}))
            elif self.criteria == COLUMN_CRITERIA_FALSE:
                return objects.filter(Q(**{self.label: False}))
            elif self.criteria == COLUMN_CRITERIA_NULL:
                return objects.filter(Q(**{self.label_with_operator('isnull'): True}))
            elif self.criteria == COLUMN_CRITERIA_NOT_NULL:
                return objects.filter(
                    Q(**{self.label_with_operator('isnull'): False})
                ).exclude(Q(**{self.label_with_operator('exact'): ''}))
        except Exception as ex:
            print("ERROR:" + str(ex))
        return objects




from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect, HttpResponseForbidden
from django.shortcuts import render
from django.urls import reverse
from django.utils.translation import gettext as _
from foledol.django.logger import log
from foledol.django.tools.field import TextField, IntegerField, BooleanField, DateField, FloatField
from foledol.django.tools.form import Form
from foledol.django.tools.handlers import confirm
from foledol.django.utils import pop_path, get_path, get_param, get_integer, error, new_context
from .grid import get_field
from .grids import grid_renumber

from ..models import Column, Grid, COLUMN_CRITERIA_SET, COLUMN_TEXT_CRITERIA_SET, COLUMN_DATE_CRITERIA_SET, \
    COLUMN_NUMBER_CRITERIA_SET, COLUMN_BOOLEAN_CRITERIA_SET


class ColumnForm(Form):
    def __init__(self, context):
        super().__init__(context, [
            TextField('label', "Libellé", min_length=1),
            IntegerField('criteria', "Critère"),
            TextField('value', "Valeur"),
            IntegerField('order', "Ordre"),
            IntegerField('width', "Largeur"),
            BooleanField('hidden', "Masquer"),
        ])


def column_form(request, column):
    if not request.user.is_staff:
        return HttpResponseForbidden()

    context = new_context()
    context['base'] = settings.DEFAULT_SPACE + '/base.html'

    action = get_path(request, context, column, 'column', space='django')

    grid_id = get_integer(request, 'grid_id', 0)
    if grid_id > 0:
        grid = Grid.objects.all().get(id=grid_id)
    context['grid_id'] = grid.id

    if column:
        context['column'] = column

    fields = grid.fields() if grid else []
    context['fields'] = fields

    context['criteria_set'] = COLUMN_CRITERIA_SET

    form = ColumnForm(context)
    form.read(request.POST, column)
    if action == 'update':
        # TODO: fix the issue on BooleanField (Read)
        form.field('hidden').value = 'hidden' in request.POST

    order = (grid.column_set.count() + 1) * 10 if grid else 10

    criteria = get_integer(request, 'criteria', column.criteria if column else 0)
    context['criteria'] = criteria

    if len(action) > 0:
        form.validate()

    if context['error']:
        context['error'] = "Veuillez corriger les erreurs ci-dessous"

    label = get_param(request, context, 'label', None)
    if label and column:
        field = get_field(column.label, fields)
        if isinstance(field, TextField):
            if criteria not in COLUMN_TEXT_CRITERIA_SET:
                context['error'] = "Le critère n'est pas compatible avec un texte"
        if isinstance(field, DateField):
            if criteria not in COLUMN_DATE_CRITERIA_SET:
                context['error'] = "Le critère n'est pas compatible avec une date"
        if isinstance(field, FloatField) or isinstance(field, IntegerField):
            if criteria not in COLUMN_NUMBER_CRITERIA_SET:
                context['error'] = "Le critère n'est pas compatible avec un nombre"
        if isinstance(field, BooleanField):
            if criteria not in COLUMN_BOOLEAN_CRITERIA_SET:
                context['error'] = "Le critère n'est pas compatible avec un booléen"

    if not action or context['error']:
        context['action'] = 'update' if column else 'create'
        return render(request, 'column.html', context)

    if action == 'create':
        column = Column(grid=grid)
    form.save(column)
    if not column.order:
        column.order = order
    column.save()
    if grid:
        grid_renumber(grid)
    log(column.id, 'column', action, request.user, form.old_values, form.new_values)

    return HttpResponseRedirect(context['back'] + '?path=' + pop_path(request))


@login_required
def column_create(request):
    return column_form(request, None)


@login_required
def column_update(request, pk):
    column = Column.objects.filter(id=pk).first()
    return column_form(request, column) if column else error(request)


@login_required
def column_delete(request, pk):
    def prepare(context):
        column = Column.objects.get(id=pk)
        context['title'] = _('delete_column_title').format(column)
        context['cancel'] = reverse('django:column_update', kwargs={'pk': pk})
        context['message'] = _('delete_column_message').format(column)
        return column

    def execute(column):
        column.delete()
        grid_renumber(column.grid)

    return confirm(request, 'delete', prepare, execute, reverse('django:columns'))



from django.conf import settings
from django.db.models import F


from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from foledol.django.tools.table import Table, TableColumn
from foledol.django.utils import get_search, paginate, new_context

from ..models import Grid


class GridTables(Table):
    def __init__(self, rows):
        super().__init__(rows, [
            TableColumn('name', "Nom"),
            TableColumn('table', "Table", method="table_as_str")
        ])
        self.update = 'django:grid_update'
        self.create = 'django:grid_create'
        self.search = True


@login_required
@staff_member_required
def grid_list(request):
    context = new_context()
    context['base'] = settings.DEFAULT_SPACE + '/base.html'

    grids = Grid.objects.all()

    search = get_search(request).strip()
    if len(search) > 0:
        grids = Grid.objects.filter(name=search)
    context['search'] = search
    context['table'] = GridTables(paginate(request, context, grids))

    return render(request, 'grids.html', context)


def grid_renumber(grid):
    order = 10
    for column in grid.column_set.all().order_by('order'):
        if column.order != order:
            column.order = order
            column.save()
        order += 10









from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from foledol.django.tools.table import Table, TableColumn
from foledol.django.utils import get_search, paginate, new_context

from ..models import Column


class ColumnTables(Table):
    def __init__(self, rows):
        super().__init__(rows, [
            TableColumn('label', "Libellé"),
            TableColumn('criteria', "Critère"),
            TableColumn('value', "Valeur")
        ])
        self.update = 'django:column_update'
        self.create = 'django:column_create'
        self.search = True


@login_required
@staff_member_required
def column_list(request):
    context = new_context()
    context['base'] = settings.DEFAULT_SPACE + '/base.html'

    columns = Column.objects.all()

    search = get_search(request).strip()
    if len(search) > 0:
        columns = Column.objects.filter(label=search)
    context['search'] = search
    context['table'] = ColumnTables(paginate(request, context, columns))

    return render(request, 'columns.html', context)



from django.contrib.auth.models import User

from .meta import get_user_or_superuser
from .models import Log, LogItem


def log(ref, model, action, user=None, old='', new='', transaction=None):
    _log = Log(user=get_user_or_superuser(user))
    _log.ref = ref
    _log.model = model
    _log.action = action
    _log.transaction = transaction
    _log.save()

    if old and new:
        diff = {}
        for key in new.keys():
            if old[key] != new[key]:
                diff[key] = (old[key], new[key])
        for key in diff.keys():
            _log_item = LogItem()
            _log_item.log = _log
            _log_item.key = key
            _log_item.old = diff[key][0]
            _log_item.new = diff[key][1]
            _log_item.save()


def logs(ref, model):
    return Log.objects.all().filter(ref=ref).filter(model=model).order_by('-date')


